//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

import SafariServices

import RxSwift
import RxCocoa
import Alamofire
import RxAlamofire

import KakaoSDKCommon
import RxKakaoSDKCommon

import KakaoSDKAuth
import RxKakaoSDKAuth

import KakaoSDKUser
import RxKakaoSDKUser

import KakaoSDKTalk
import RxKakaoSDKTalk

import KakaoSDKStory
import RxKakaoSDKStory

import KakaoSDKLink
import RxKakaoSDKLink

import KakaoSDKNavi
import KakaoSDKTemplate

import RxDataSources

// MARK: - tableview helper =============================================================================================

public class Row {
    public var title: String
    public var action: (() -> Void)?
    public var apiResult: ApiResult
    public init(title: String, action: (() -> Void)?, result: ApiResult = .Ready) {
        self.title = title
        self.action = action
        self.apiResult = result
    }
}

public class SectionOfRow : SectionModelType {
    public typealias Item = Row
    
    public var header: String
    public var items: [Item]
    public init(header:String, items:[Item]) {
        self.header = header
        self.items = items
    }
    
    required public convenience init(original: SectionOfRow, items: [Item]) {
        self.init(header: original.header, items: original.items)
        self.items = items
    }
}

// MARK: tableview helper =============================================================================================
// MARK: -

extension RxSDKSampleViewController {
    private func updateResultIndicator(cell:UITableViewCell?, row:Row) {
        
        switch(row.apiResult) {
        case .Ready: cell?.backgroundColor = UIColor.white
        case .Success: cell?.backgroundColor = UIColor.green
        case .Fail: cell?.backgroundColor = UIColor.red
        }
    }
    
    private func updateResultIndicator() {
        if let resultIndicator = self.resultIndicator {
            let cell = self.tableView.cellForRow(at: resultIndicator.0)
            self.updateResultIndicator(cell: cell, row: resultIndicator.1)
        }
    }
    
    private func success(_ object: Any? = nil, updateIndicator: Bool = true) {
        if let object = object {
            print("\(String(describing:object))")
        }
        
        if (updateIndicator) {
            self.resultIndicator?.1.apiResult = .Success
            self.updateResultIndicator()
        }
    }
    
    
    private func errorHandler(error:Error?, updateIndicator: Bool = true) {
        guard error != nil else { return }
        if (updateIndicator) {
            self.resultIndicator?.1.apiResult = .Fail
            self.updateResultIndicator()
        }
        
        if let sdkError = error as? SdkError {
            print("[sdk error handling start]------------------------------------------------------------------------------")
            
            switch(sdkError) {
            case .ApiFailed:
                let apiError = sdkError.getApiError()
                switch(apiError.reason) {
                default:
                    print("ApiError 서드 개발자 에러처리 필요.")
                }
                break
            case .AuthFailed:
                let authError = sdkError.getAuthError()
                switch(authError.reason) {
                default:
                    print("AuthError 서드 개발자 에러처리 필요.")
                }
                break
            case .ClientFailed:
                break
            case .GeneralFailed:
                break
            }
            print("[sdk error handling end]--------------------------------------------------------------------------------\n")
        }
        else {
            print("[not sdk error]------------------------------------------------------------------------------")
            print("\(String(describing: error))")
        }
    }
}


public class RxSDKSampleViewController: SDKSampleBaseViewController {
    let disposeBag = DisposeBag()
    
    lazy public var dataSource = RxTableViewSectionedReloadDataSource<SectionOfRow>(
        configureCell: { (dataSource, tableView, indexPath, item) in
            let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
            
            self.updateResultIndicator(cell:cell, row:item)
            
//            cell.selectionStyle = UITableViewCell.SelectionStyle.none
            cell.textLabel?.adjustsFontSizeToFitWidth = true
            cell.textLabel?.text = "\(item.title)"
            cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
            cell.textLabel?.textColor = UIColor.black
            return cell
        })
    
    var safariViewController : SFSafariViewController?
    var authenticationSession : SFAuthenticationSession?
    
    
    //for ui
    var resultIndicator : (IndexPath, Row)?
    
    override public func loadView() {
        super.loadView()
        
        setupViews()
    }
    
    override public func viewDidLoad() {
        super.viewDidLoad()
        
        setMenu()
        
        bindViews()
        
        subscribe2()
        
        //COMMENT: 릴리즈모드에서도 subscribe 가능.
//        SdkLog.logger.subscribe{print($0)}
//        .disposed(by: self.disposeBag)
    }
    
    override public func setupViews() {
        super.setupViews()
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "Cell")
        
//        self.navigationItem.rightBarButtonItems = [UIBarButtonItem(title: "Log", style: UIBarButtonItem.Style.plain, target: self, action: #selector(RxSDKSampleViewController.log(sender:)))]
    }
    
//    @objc func log(sender:UIBarButtonItem) {
//        print("logbutton clicked")
//    }
    
    private func bindViews() {
        self.dataSource.titleForHeaderInSection = { dataSource, index in
            return dataSource.sectionModels[index].header
        }
        
        Observable.just(self.sections!)
            .bind(to: tableView.rx.items(dataSource: dataSource))
            .disposed(by: disposeBag)
    }
    
//    func subscribe() {
//        self.tableView.rx.modelSelected(Row.self).subscribe(onNext: { row in
//            if let action = row.action {
//                action()
//            }
//        })
//            .disposed(by: disposeBag)
//    }
    
    func subscribe2() {
        Observable
            .zip(self.tableView.rx.itemSelected, self.tableView.rx.modelSelected(Row.self),
                 resultSelector: { indexPath, row in
                    if let action = row.action {
                        action()
                        
                        self.resultIndicator = (indexPath, row)
                        
                        DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
                           self.tableView.deselectRow(at: indexPath, animated: false)
                        }
                    }
            })
            .subscribe()
            .disposed(by: self.disposeBag)
    }
    
    public func setTitle(title:String = "Rx SDK Sample", args: String? = nil) {
        self.navigationItem.title = (args != nil) ? "\(title)(\(args!))" : "\(title)"
    }
    
    func extractPickerElements(friends:Friends<Friend>) -> [(URL?, String?, Any)]? {
        return friends.elements?.map({ (friend) -> (URL?, String?, Any) in return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid) })
    }
    
    func setMenu() {
        self.sections = [
            // MARK: - section auth start =============================================================================================
            SectionOfRow(header: "AUTH", items: [
                Row(title: "loginWithKakaoTalk()", action: {
                    if (AuthApi.isKakaoTalkLoginAvailable()) {
                        AuthApi.shared.rx.loginWithKakaoTalk()
                            .subscribe(onNext:{ (oauthToken) in
                            self.success()
                            
                            let _ = oauthToken
                            //do something
                                
                        })
                        .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title: "loginWithKakaoAccount() - iOS 11.0~", action: {
                    AuthApi.shared.rx.loginWithKakaoAccount()
                        .subscribe(onNext:{ (oauthToken) in
                            self.success()
                            
                            let _ = oauthToken
                            //do something
                            
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "===== kakaotalk login(universal link) - under developing", action: {
                    print("under developing")
                }),
            ]),
            // section auth end =============================================================================================
            
            // MARK: - section user start =============================================================================================
            SectionOfRow(header: "USER", items: [
                Row(title:"logout()", action: {
                    UserApi.shared.rx.logout()
                        .subscribe(onCompleted:{
                            self.success("onCompleted:")
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"unlink()", action: {
                    UserApi.shared.rx.unlink()
                        .subscribe(onCompleted:{
                            self.success("onCompleted:")
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "me()", action: {
                    UserApi.shared.rx.me()
                        .subscribe (onSuccess:{ user in
                            self.success()
                            
                            let _ = user
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "me() dynamic agreement", action: {
                    UserApi.shared.rx.me()
                        .map({ (user) -> User in
                            var scopes = [String]()
                            
                            if (user.kakaoAccount?.emailNeedsAgreement == true) { scopes.append("account_email") }
                            if (user.kakaoAccount?.birthdayNeedsAgreement == true) { scopes.append("birthday") }
                            if (user.kakaoAccount?.birthyearNeedsAgreement == true) { scopes.append("birthyear") }
                            if (user.kakaoAccount?.ciNeedsAgreement == true) { scopes.append("account_ci") }
                            if (user.kakaoAccount?.legalNameNeedsAgreement == true) { scopes.append("legal_name") }
                            if (user.kakaoAccount?.legalBirthDateNeedsAgreement == true) { scopes.append("legal_birth_date") }
                            if (user.kakaoAccount?.legalGenderNeedsAgreement == true) { scopes.append("legal_gender") }
                            
                            if (user.kakaoAccount?.phoneNumberNeedsAgreement == true) { scopes.append("phone_number") }
                            if (user.kakaoAccount?.profileNeedsAgreement == true) { scopes.append("profile") }
                            if (user.kakaoAccount?.ageRangeNeedsAgreement == true) { scopes.append("age_range") }
                            
                            if (scopes.count > 0) {
                                throw SdkError(scopes:scopes)
                            }
                            else {
                                return user
                            }
                        })
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe(onSuccess:{ ( user ) in
                            self.success()
                            
                            let _ = user
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"updateProfile()", action: {
                    UserApi.shared.rx.updateProfile(properties:["nickname":"update_profile_test"])
                        .subscribe(onCompleted:{
                            self.success()
                            
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"accessTokenInfo()", action: {
                    UserApi.shared.rx.accessTokenInfo()
                        .subscribe(onSuccess:{ (accessTokenInfo) in
                            self.success()
                            
                            let _ = accessTokenInfo
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"shippingAddresses(fromUpdatedAt:pageSize:)", action: {
                    UserApi.shared.rx.shippingAddresses()
                        .subscribe(onSuccess:{ (shippingAddress) in
                            self.success()
                            
                            let _ = shippingAddress
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"serviceTerms()", action: {
                    UserApi.shared.rx.serviceTerms()
                        .subscribe(onSuccess:{ (serviceTerms) in
                            self.success()
                            
                            let _ = serviceTerms
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                })
            ]),
            // section user end =============================================================================================
            
            // MARK: - section stroy start =============================================================================================
            SectionOfRow(header: "STORY", items: [
                Row(title: "isStoryUser()", action: {
                    StoryApi.shared.rx.isStoryUser()
                        .subscribe (onSuccess:{ (isStoryUser) in
                            self.success()
                            
                            let _ = isStoryUser
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "profile()", action: {
                    StoryApi.shared.rx.profile()
                        .subscribe (onSuccess:{ (profile) in
                            self.success()
                            
                            let _ = profile
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "stories()", action: {
                    StoryApi.shared.rx.stories()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (stories) in
                            self.success()
                            
                            let _ = stories
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "story(id:)", action: {
                    StoryApi.shared.rx.stories()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .flatMap({ (stories) -> Single<Story> in
                            //first element get example
                            if let first = stories?.first {
                                return StoryApi.shared.rx.story(id:first.id)
                            }
                            else {
                                throw SdkError(reason:.Unknown, message: "not exist your story.")
                            }
                        })
                        .subscribe (onSuccess:{ (story) in
                            self.success()
                            
                            let _ = story
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "linkInfo(url:)", action: {
                    StoryApi.shared.rx.linkInfo(url: URL(string: "https://www.youtube.com/watch?v=naytyuVj75I")!)
                        .subscribe (onSuccess:{ (linkInfo) in
                            self.success()
                            
                            let _ = linkInfo
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "postNote(content:)", action: {
                    StoryApi.shared.rx.postNote(content: "This cafe is really awesome!")
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (postId) in
                            self.success()
                            
                            let _ = postId
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "postLink(linkInfo:)", action: {
                    StoryApi.shared.rx.linkInfo(url: URL(string: "https://www.youtube.com/watch?v=naytyuVj75I")!)
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .flatMap({ linkInfo in
                            StoryApi.shared.rx.postLink(linkInfo: linkInfo)
                                .retryWhen(Auth.shared.rx.dynamicAgreement())
                        })
                        .subscribe (onSuccess:{ (postId) in
                            self.success()
                            
                            let _ = postId
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "postPhoto(imagePaths:)", action: {
                    StoryApi.shared.rx.upload([UIImage(named:"sample1"), UIImage(named:"sample2")])
                        .flatMap({ (imagePaths) -> Single<String> in
                            
                            guard let imagePaths = imagePaths else {
                                throw (NSError(domain:"com.kakao.sdk.KakaoSample", code:0, userInfo:[ NSLocalizedDescriptionKey: "Upload is not complete."]) as Error)
                            }
                            return StoryApi.shared.rx.postPhoto(imagePaths:imagePaths)
                                .retryWhen(Auth.shared.rx.dynamicAgreement())
                        })
                        .subscribe (onSuccess:{ (postId) in
                            self.success()
                            
                            let _ = postId
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "uploadMulti([UIImage])", action: {
                    StoryApi.shared.rx.upload([UIImage(named:"sample1"), UIImage(named:"sample2")])
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (postIds) in
                            self.success()
                            
                            let _ = postIds
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "delete(_ id:) - first element delete", action: {
                    StoryApi.shared.rx.stories()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .flatMapCompletable({ (stories) -> Completable in
                            //first element delete example
                            if let first = stories?.first {
                                return StoryApi.shared.rx.delete(first.id)
                            }
                            else {
                                throw SdkError(reason:.Unknown, message: "not exist your story.")
                            }
                        })
                        .subscribe (onCompleted:{
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
            ]),
            // section story end =============================================================================================
            
            // MARK: - section talk start =============================================================================================
            //MARK: - send memo
            SectionOfRow(header: "TALK", items: [
                Row(title: "profile()", action: {
                    TalkApi.shared.rx.profile()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (profile) in
                            self.success()
                            
                            let _ = profile
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "friends()", action: {
                    TalkApi.shared.rx.friends()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (friends) in
                            self.success()
                            
                            let _ = friends
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "friends(offset:limit:order:)", action: {
                    //TODO:
                    //적합하지 않은 예제.
                    
//                    Single.zip(
//                        TalkApi.shared.rx.friends(offset: 0, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.dynamicAgreement()),
//                        TalkApi.shared.rx.friends(offset: 1, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.dynamicAgreement()),
//                        TalkApi.shared.rx.friends(offset: 2, limit: 1, order: .Desc).retryWhen(Auth.shared.rx.dynamicAgreement()),
//                        resultSelector: { first, second, third in
//                            print(" ---->first:\(first)\n ---->second:\(second)\n ---->third:\(third)")
//                            return Friends<Friend>.merge([first, second, third])
//                        })
//                        .subscribe (onSuccess:{
//                            //TODO: 뭔가 좀 이상 다시한번 보기.
//
//                            self.success($0)
////                            let _ = friends
//
//
//                        }, onError: { error in
//                            self.errorHandler(error: error)
//                        })
//                    .disposed(by: self.disposeBag)
                }),
                
                Row(title: "channels()", action: {
                    TalkApi.shared.rx.channels()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onSuccess:{ (channels) in
                            self.success()
                            
                            let _ = channels
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendDefaultMemo(templatable:)", action: {
                    let title = "카카오톡 링크 (메모)"
                    let description = "나에게 메모 티폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    //TODO: 외부에 오픈하는거니까 디코더를 직접 쓰지 않는 메소드를 하나 만들자.
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        print("Template URL decoding test: \(templatable.content.imageUrl.absoluteString)")
                        TalkApi.shared.rx.sendDefaultMemo(templatable:templatable)
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe (onCompleted:{
                                
                                self.success()
                                //do something
                                
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title: "sendCustomMemo(templateId:)", action: {
                    TalkApi.shared.rx.sendCustomMemo(templateId: AppDelegate.templateId().message)
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onCompleted:{
                            
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendCustomMemo(templateId:templateArgs:)", action: {
                    TalkApi.shared.rx.sendCustomMemo(templateId: AppDelegate.templateId().message, templateArgs: ["msg":"메세지 파라미터 테스트입니다."])
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onCompleted:{
                            
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendScrapMemo(requestUrl:)", action: {
                    TalkApi.shared.rx.sendScrapMemo(requestUrl: "https://developers.kakao.com")
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe (onCompleted:{
                            
                            self.success()
                            //do something
                            
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                
                //MARK: - send message
                Row(title: "sendDefaultMessage(templatable:receiverUuids:) - feed", action: {
                    
                    let title = "카카오톡 링크 (메세지)"
                    let description = "친구에게보내기 티폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                    """
                        {
                            "object_type": "feed",
                            "content": {
                                "title": "\(title)",
                                "description": "\(description)",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "http://dev.kakao.com",
                                    "mobile_web_url": "http://dev.kakao.com"
                                }
                            },
                            "social": {
                                "like_count": 100,
                                "comment_count": 200
                            },
                            "button_title": "바로 확인"
                        }
                        """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        TalkApi.shared.rx.friends()
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                                                                    .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                    .subscribe(onSuccess:{ (messageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = messageSendResult
                                                                                        

                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                
//             =================================
                
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - list", action: {
                    
                    let headerTitle = "카카오 디벨로퍼"
                    let listTemplateJsonStringData =
                    """
                    {
                        "object_type": "list",
                        "header_title": "\(headerTitle)",
                        "header_link": {
                            "web_url": "https://dev.kakao.com",
                            "mobile_web_url": "https://dev.kakao.com"
                        },
                        "contents": [
                            {
                                "title": "API 소개",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://dev.kakao.com/features/platform",
                                    "mobile_web_url": "https://dev.kakao.com/features/platform"
                                }
                            },
                            {
                                "title": "개발자 가이드",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://dev.kakao.com/docs",
                                    "mobile_web_url": "https://dev.kakao.com/docs"
                                }
                            },
                            {
                                "title": "개발자 포럼",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://devtalk.kakao.com",
                                    "mobile_web_url": "https://devtalk.kakao.com"
                                }
                            }
                        ],
                        "button_title": "웹에서 보기"
                    }
                    """.data(using: .utf8)!
                    
//                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
//                        LinkApi.shared.rx.defaultLink(templatable:templatable)
//                            .subscribe(onSuccess: { (linkResult) in
//                                self.success(linkResult)
//
//                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
//                            }, onError: {error in
//                                self.errorHandler(error: error)
//                            })
//                            .disposed(by: self.disposeBag)
//                    }
                    
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
                        TalkApi.shared.rx.friends()
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                                                                    .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                    .subscribe(onSuccess:{ (messageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = messageSendResult
                                                                                        
                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - location", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 Location"
                    let locationTemplateJsonStringData =
                    """
                    {
                        "object_type": "location",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인",
                        "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 7층",
                        "address_title": "카카오"
                    }
                    """.data(using: .utf8)!
                    
//                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
//                        LinkApi.shared.rx.defaultLink(templatable:templatable)
//                            .subscribe(onSuccess: { (linkResult) in
//                                self.success(linkResult)
//
//                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
//                            }, onError: {error in
//                                self.errorHandler(error: error)
//                            })
//                            .disposed(by: self.disposeBag)
//                    }
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                        TalkApi.shared.rx.friends()
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                                                                    .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                    .subscribe(onSuccess:{ (messageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = messageSendResult
                                                                                        
                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - commerce", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 Commerce"
                    let commerceTemplateJsonStringData =
                    """
                    {
                        "object_type": "commerce",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "commerce": {
                            "regular_price": 10000
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
//                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
//                        LinkApi.shared.rx.defaultLink(templatable:templatable)
//                            .subscribe(onSuccess: { (linkResult) in
//                                self.success(linkResult)
//
//                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
//                            }, onError: {error in
//                                self.errorHandler(error: error)
//                            })
//                            .disposed(by: self.disposeBag)
//                    }
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                        TalkApi.shared.rx.friends()
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                                                                    .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                    .subscribe(onSuccess:{ (messageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = messageSendResult
                                                                                        
                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"sendDefaultMessage(templatable:receiverUuids:) - text", action: {
                    let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                    let textTemplateJsonStringData =
                    """
                    {
                        "object_type": "text",
                        "text": "\(text)",
                        "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
//                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
//                        LinkApi.shared.rx.defaultLink(templatable:templatable)
//                            .subscribe(onSuccess: { (linkResult) in
//                                self.success(linkResult)
//
//                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
//                            }, onError: {error in
//                                self.errorHandler(error: error)
//                            })
//                            .disposed(by: self.disposeBag)
//                    }
//                    else {
//                        print("templatable is nil")
//                    }
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                        TalkApi.shared.rx.friends()
                            .retryWhen(Auth.shared.rx.dynamicAgreement())
                            .subscribe(onSuccess: { (friends) in
                                self.success(friends, updateIndicator: false)
                                
                                if let elements = self.extractPickerElements(friends: friends) {
                                    let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                                                                let friendUuids = selectedIds as! [String]
                                                                                TalkApi.shared.rx.sendDefaultMessage(templatable:templatable, receiverUuids:friendUuids)
                                                                                    .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                    .subscribe(onSuccess:{ (messageSendResult) in
                                                                                        self.success()
                                                                                        
                                                                                        //do something
                                                                                        let _ = messageSendResult
                                                                                    
                                                                                    }, onError: {error in
                                                                                        self.errorHandler(error: error)
                                                                                    })
                                                                                    .disposed(by: self.disposeBag)
                                                                                
                                    })
                                    self.navigationController?.pushViewController(vc, animated: true)
                                }
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title: "sendCustomMessage(templateId:templateArgs:receiverUuids:)", action: {
                    TalkApi.shared.rx.friends()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendCustomMessage(templateId:AppDelegate.templateId().message, templateArgs:["title":"제목입니다.", "description":"설명입니다."], receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                .subscribe(onSuccess:{ (messageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = messageSendResult

                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "sendScrapMessage(requestUrl:receiverUuids:)", action: {
                    TalkApi.shared.rx.friends()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendScrapMessage(requestUrl:"https://developers.kakao.com", receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                .subscribe(onSuccess:{ (messageSendResult) in
                                                                                    self.success()
                                                                                    
                                                                                    //do something
                                                                                    let _ = messageSendResult
      
                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    
                }),
                Row(title: "sendScrapMessage(requestUrl:templateId:receiverUuids:)", action: {
                    TalkApi.shared.rx.friends()
                        .retryWhen(Auth.shared.rx.dynamicAgreement())
                        .subscribe(onSuccess: { (friends) in
                            self.success(friends, updateIndicator: false)
                            
                            if let elements = self.extractPickerElements(friends: friends) {
                                let vc = RxSDKSamplePickerViewController(title: "Friends",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let friendUuids = selectedIds as! [String]
                                                                            TalkApi.shared.rx.sendScrapMessage(requestUrl:"https://developers.kakao.com", templateId:AppDelegate.templateId().message, receiverUuids:friendUuids)
                                                                                .retryWhen(Auth.shared.rx.dynamicAgreement())
                                                                                .subscribe(onSuccess:{ (messageSendResult) in
                                                                                   self.success()
                                                                                   
                                                                                   //do something
                                                                                   let _ = messageSendResult
                                                                                    
                                                                                }, onError: {error in
                                                                                    self.errorHandler(error: error)
                                                                                })
                                                                                .disposed(by: self.disposeBag)
                                                                            
                                })
                                self.navigationController?.pushViewController(vc, animated: true)
                            }
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                    
                }),
                Row(title: "addChannel(channelPublicId:) - web", action: {
                    self.safariViewController = SFSafariViewController(url: TalkApi.shared.makeUrlForAddChannel(channelPublicId:"_ZeUTxl")!)
                    self.safariViewController?.modalTransitionStyle = .crossDissolve
                    self.safariViewController?.modalPresentationStyle = .overCurrentContext
                    self.present(self.safariViewController!, animated: true) {
                        print("add plusfriend present success")
                    }
                }),
                Row(title: "chat(channelPublicId:) - web", action: {
                    self.safariViewController = SFSafariViewController(url: TalkApi.shared.makeUrlForChannelChat(channelPublicId:"_ZeUTxl")!)
                    self.safariViewController?.modalTransitionStyle = .crossDissolve
                    self.safariViewController?.modalPresentationStyle = .overCurrentContext
                    self.present(self.safariViewController!, animated: true) {
                        print("chat present success")
                    }
                })
                
            ]),
            // section talk end =============================================================================================
            
            // MARK: - section kakaolink start =============================================================================================
            SectionOfRow(header: "KAKAOLINK", items: [
                // TODO: canOpenURL 함께 사용하는 예제 추가
                Row(title: "defaultLink(templatable:) - webSharer", action: {
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        if let url = LinkApi.shared.makeSharerUrlforDefaultLink(templatable: templatable) {
                            print("=============================sharer default url: \(url)")
                            // TODO: 유니버셜링크로 적용 해봐요 https://{phase-}talk-apps.kakao.com/scheme/{encodedLinkUrl}&web={encodedSharerUrl}
                            // 장점:
                            //   - 개발가이드에서 LSApplicationQueriesSchemes kakaolink, kakaotalk-5.9.7 제외 가능
                            //     - kakaolink, kakaotalk-5.9.7 모두 제거하는 경우: 현재 카톡 최소지원버전이 5.9.0이고 linkver=4.0 지원 버전이 5.9.8이므로 에러케이스가 극소수 발생
                            //     - kakaolink만 제거하는 경우: 이슈 없음. 조금 아쉬움
                            //   - 스킴 오등록 대응
                            self.authenticationSession = SFAuthenticationSession(url: url, callbackURLScheme: nil, completionHandler: { [weak self] (_,_) in
                                self?.authenticationSession = nil
                            })
                            self.authenticationSession?.start()
                        }
                    }
                }),
                Row(title: "scrapLink(requestUrl:) - webSharer", action: {
                    if let url = LinkApi.shared.makeSharerUrlforScrapLink(requestUrl: "https://developers.kakao.com") {
                        print("=============================sharer scrap url: \(url)")
                        // TODO: 유니버셜링크로 적용 해봐요 https://{phase-}talk-apps.kakao.com/scheme/{encodedLinkUrl}&web={encodedSharerUrl}
                        self.authenticationSession = SFAuthenticationSession(url: url, callbackURLScheme: nil, completionHandler: { [weak self] (_,_) in
                            self?.authenticationSession = nil
                        })
                        self.authenticationSession?.start()
                    }
                }),
                Row(title: "scrapLink(requestUrl:templateId:) - webSharer", action: {
                    if let url = LinkApi.shared.makeSharerUrlforScrapLink(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link ) {
                        print("=============================sharer scrap url: \(url)")
                        // TODO: 유니버셜링크로 적용 해봐요 https://{phase-}talk-apps.kakao.com/scheme/{encodedLinkUrl}&web={encodedSharerUrl}
                        self.authenticationSession = SFAuthenticationSession(url: url, callbackURLScheme: nil, completionHandler: { [weak self] (_,_) in
                            self?.authenticationSession = nil
                        })
                        self.authenticationSession?.start()
                    }
                }),
                Row(title: "customLink(templateId:,templateArgs:) - webSharer", action: {
                    if let url = LinkApi.shared.makeSharerUrlforCustomLink(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."]) {
                        print("=============================sharer custom url: \(url)")
                        // TODO: 유니버셜링크로 적용 해봐요 https://{phase-}talk-apps.kakao.com/scheme/{encodedLinkUrl}&web={encodedSharerUrl}
                        self.authenticationSession = SFAuthenticationSession(url: url, callbackURLScheme: nil, completionHandler: { [weak self] (_,_) in
                            self?.authenticationSession = nil
                        })
                        self.authenticationSession?.start()
                    }
                }),
                Row(title: "scrapLink(requestUrl:)", action: {
                    LinkApi.shared.rx.scrapLink(requestUrl: "https://developers.kakao.com")
                        .subscribe(onSuccess: { (linkResult) in
                            self.success()
                            
                            UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "scrapLink(requestUrl:templateId:)", action: {
                    LinkApi.shared.rx.scrapLink(requestUrl: "https://developers.kakao.com", templateId: AppDelegate.templateId().link)
                        .subscribe(onSuccess: { (linkResult) in
                            self.success()
                            
                            UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title: "customLink(templateId:templateArgs:)", action: {
                    LinkApi.shared.rx.customLink(templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."])
                        .subscribe(onSuccess: { (linkResult) in
                            self.success()
                            
                            UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                        }, onError: {error in
                            self.errorHandler(error: error)
                        })
                        .disposed(by: self.disposeBag)
                }),
                Row(title:"defaultLink(templatable:)", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templatable:templatable)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"defaultLink(templateObject:) using json", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 FEED"
                    
                    let feedTemplateJsonStringData =
                    """
                    {
                        "object_type": "feed",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
                    //jsonObject
                    if let feedTemplateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templateObject:feedTemplateJsonObject)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"defaultLink(templateObject:) using constructor", action: {
                    
                    let title = "카카오톡 링크 using constructor"
                    let description = "디폴트 템플릿 FEED using constructor"
                    
                    let button1 = Button(title: "button1", link: Link(webUrl: URL(string:"http://dev.kakao.com"), mobileWebUrl: URL(string:"http://dev.kakao.com")))
                    let button2 = Button(title: "button2", link: Link(webUrl: URL(string:"http://dev.kakao.com"), mobileWebUrl: URL(string:"http://dev.kakao.com")))
                    
                    let social = Social(likeCount: 100, commentCount: 200)
                    let link = Link(webUrl: URL(string:"http://dev.kakao.com"), mobileWebUrl: URL(string:"http://dev.kakao.com"))
                    let content = Content(title: title,  imageUrl: URL(string:"http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png")!, description: description, link: link)
                    let feedTemplate = FeedTemplate(content: content, social: social, buttons: [button1, button2])
                    
                    //encode 는 이렇게 한다.
                    if let feedTemplateJsonData = (try? SdkJSONEncoder.custom.encode(feedTemplate)) {
                    
                    //jsonObject
                        if let templateJsonObject = SdkUtils.toJsonObject(feedTemplateJsonData) {
                            LinkApi.shared.rx.defaultLink(templateObject:templateJsonObject)
                                .subscribe(onSuccess: { (linkResult) in
                                    self.success()
                                    
                                    UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                                }, onError: {error in
                                    self.errorHandler(error: error)
                                })
                                .disposed(by: self.disposeBag)
                        }
                    }
                }),
                Row(title:"defaultLink(templatable:) - list", action: {
                    
                    let headerTitle = "카카오 디벨로퍼"
                    let listTemplateJsonStringData =
                    """
                    {
                        "object_type": "list",
                        "header_title": "\(headerTitle)",
                        "header_link": {
                            "web_url": "https://dev.kakao.com",
                            "mobile_web_url": "https://dev.kakao.com"
                        },
                        "contents": [
                            {
                                "title": "API 소개",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://dev.kakao.com/features/platform",
                                    "mobile_web_url": "https://dev.kakao.com/features/platform"
                                }
                            },
                            {
                                "title": "개발자 가이드",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://dev.kakao.com/docs",
                                    "mobile_web_url": "https://dev.kakao.com/docs"
                                }
                            },
                            {
                                "title": "개발자 포럼",
                                "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                                "link": {
                                    "web_url": "https://devtalk.kakao.com",
                                    "mobile_web_url": "https://devtalk.kakao.com"
                                }
                            }
                        ],
                        "button_title": "웹에서 보기"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(ListTemplate.self, from: listTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templatable:templatable)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"defaultLink(templatable:) - location", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 Location"
                    let locationTemplateJsonStringData =
                    """
                    {
                        "object_type": "location",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "social": {
                            "like_count": 100,
                            "comment_count": 200
                        },
                        "button_title": "바로 확인",
                        "address": "경기 성남시 분당구 판교역로 235 에이치스퀘어 N동 7층",
                        "address_title": "카카오"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(LocationTemplate.self, from: locationTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templatable:templatable)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    
                }),
                Row(title:"defaultLink(templatable:) - commerce", action: {
                    
                    let title = "카카오톡 링크"
                    let description = "디폴트 템플릿 Commerce"
                    let commerceTemplateJsonStringData =
                    """
                    {
                        "object_type": "commerce",
                        "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                                "web_url": "http://dev.kakao.com",
                                "mobile_web_url": "http://dev.kakao.com"
                            }
                        },
                        "commerce": {
                            "regular_price": 10000
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(CommerceTemplate.self, from: commerceTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templatable:templatable)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                }),
                Row(title:"defaultLink(templatable:) - text", action: {
                    let text = "카카오링크는 카카오 플랫폼 서비스의 대표 기능으로써 사용자의 모바일 기기에 설치된 카카오 플랫폼과 연동하여 다양한 기능을 실행할 수 있습니다.\\n\\n현재 이용할 수 있는 카카오링크는 다음과 같습니다.\\n카카오톡링크\\n카카오톡을 실행하여 사용자가 선택한 채팅방으로 메시지를 전송합니다.\\n카카오스토리링크\\n카카오스토리 글쓰기 화면으로 연결합니다."
                    let textTemplateJsonStringData =
                    """
                    {
                        "object_type": "text",
                        "text": "\(text)",
                        "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                        },
                        "button_title": "바로 확인"
                    }
                    """.data(using: .utf8)!
                    
                    if let templatable = try? SdkJSONDecoder.custom.decode(TextTemplate.self, from: textTemplateJsonStringData) {
                        LinkApi.shared.rx.defaultLink(templatable:templatable)
                            .subscribe(onSuccess: { (linkResult) in
                                self.success()
                                
                                UIApplication.shared.open(linkResult.url, options: [:], completionHandler: nil)
                            }, onError: {error in
                                self.errorHandler(error: error)
                            })
                            .disposed(by: self.disposeBag)
                    }
                    else {
                        print("templatable is nil")
                    }
                }),
                
            ]),
            // section kakaolink end =============================================================================================
            // MARK: - section kakao navi start =============================================================================================
            SectionOfRow(header: "KAKAONAVI", items: [
                Row(title: "share(destination:)", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: 321286, y: 533707)
                    guard let shareUrl = NaviApi.shared.shareUrl(destination: destination) else {
                        return
                    }
                    UIApplication.shared.open(shareUrl, options: [:], completionHandler: nil)
                }),
                Row(title: "navigate(destination:viaList:)", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: 321286, y: 533707)
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: 321525, y: 532951)]
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination, viaList: viaList) else {
                        return
                    }
                    UIApplication.shared.open(navigateUrl, options: [:], completionHandler: nil)
                }),
                Row(title: "navigate(destination:viaList:) - web", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: 321286, y: 533707)
                    let viaList = [NaviLocation(name: "판교역 1번출구", x: 321525, y: 532951)]
                    self.safariViewController = SFSafariViewController(url: NaviApi.shared.webNavigateUrl(destination: destination, viaList:viaList)!)
                    self.safariViewController?.modalTransitionStyle = .crossDissolve
                    self.safariViewController?.modalPresentationStyle = .overCurrentContext
                    self.present(self.safariViewController!, animated: true) {
                        print("웹안내 present success")
                    }
                }),
                Row(title: "navigate - fallback", action: {
                    let destination = NaviLocation(name: "카카오판교오피스", x: 321286, y: 533707)
                    guard let navigateUrl = NaviApi.shared.navigateUrl(destination: destination) else {
                        return
                    }
                    
                    // canOpenURL을 사용하려면 앱의 설정파일 LSApplicationQueriesSchemes에 "kakaonavi-sdk" 스킴이 등록되어야 합니다. 샘플앱의 info.plist를 참고해주세요.
                    if UIApplication.shared.canOpenURL(navigateUrl) {
                        UIApplication.shared.open(navigateUrl, options: [:], completionHandler: { (success) in
                            if success {
                                print("장소 공유 성공")
                            } else {
                                print("장소 공유가 취소되었습니다.")
                            }
                        })
                    } else {
                        print("카카오내비 앱이 설치되지 않아 웹페이지로 엽니다.")
                        guard let webNavigateUrl = NaviApi.shared.webNavigateUrl(destination: destination) else {
                            return
                        }
                        UIApplication.shared.open(webNavigateUrl, options: [:], completionHandler: nil)
                    }
                })
            ]),
            // section kakao navi end =============================================================================================
            
            // MARK: - section etc start =============================================================================================
            SectionOfRow(header: "ETC", items: [
                Row(title: "Set custom token_manager ", action: {
                    print(">> Set custom token manager ")
                    Auth.shared.setTokenManager(TestCustomTokenManager())
                }),
                Row(title: "Check custom token_manager token ", action: {
                    var found = false
                    for (key, value) in UserDefaults.standard.dictionaryRepresentation() {
                        if (key == TestCustomTokenManager.tokenKey) {
                            print(">> custom token_manager token: \(String(describing:value))\n")
                            found = true
                            break
                        }
                    }
                    
                    if (found == false) {
                        print(">> not exist custom token_manager token")
                    }
                }),
                Row(title: "Set default token_manager ", action: {
                    print(">> Set default token manager ")
                    Auth.shared.setTokenManager()
                })
            ])
            // section etc start =============================================================================================
        ]
        
    }

}


class TestCustomTokenManager : TokenManagable {
    public static let tokenKey = "test_token_key"
    
    
    func getToken() -> OAuthToken? {
        print(">> TestCustomTokenManager.getToken() called.")
        
        if let data = UserDefaults.standard.data(forKey: TestCustomTokenManager.tokenKey) {
            let token = try? JSONDecoder().decode(OAuthToken.self, from:data)
            print(">> loaded token: \(String(describing:token))\n\n")
            
            return token
        }
        return nil
    }
    
    func setToken(_ token: OAuthToken?) {
        print(">> TestCustomTokenManager.setToke(_) called. token: \(String(describing:token)) ")
        
        if let data = try? JSONEncoder().encode(token) {
            print(">> 토큰 암호화를 권장합니다.\n\n")
            
            UserDefaults.standard.set(data, forKey:TestCustomTokenManager.tokenKey)
            UserDefaults.standard.synchronize()
        }
    }
    
    func deleteToken() {
        print(">> TestCustomTokenManager.deleteToken() called.")
        
        UserDefaults.standard.removeObject(forKey: TestCustomTokenManager.tokenKey)
    }
}
